local E, L, V, P, G = unpack(ElvUI) --Import: Engine, Locales, PrivateDB, ProfileDB, GlobalDB
local GBA = E:NewModule("GBA", "AceHook-3.0", "AceEvent-3.0", "AceTimer-3.0"); --Create a plugin within ElvUI and adopt AceHook-3.0, AceEvent-3.0 and AceTimer-3.0. We can make use of these later.
local EP = LibStub("LibElvUIPlugin-1.0") --We can use this to automatically insert our GUI tables when ElvUI_Config is loaded.
local addonName, addonTable = ... --See http://www.wowinterface.com/forums/showthread.php?t=51502&p=304704&postcount=2
local addonVersion = GetAddOnMetadata(addonName, "Version")

-- **DEBUG FUNCTIONS**
function GBA:DumpTable(table)
    for key, value in pairs(table) do
        print(key, ": ", value)
    end
end


-- **DEFAULT OPTIONS**
P["GBA"] = {
    ["CombatDamageText"] = true,
    ["CombatHealingText"] = true,
    ["WorldTextScale"] = 1,
    
    ["SpellQueueWindow"] = 200,
    ["WeatherDensity"] = 3,
    ["FfxGlow"] = true,

    ["AutoFillDelete"] = true,
    ["AutoGearCompare"] = true,
    ["Fanhexie"] = true,
    ["ProfanityFilter"] = false,
    ["EnhancedReputation"] = true,
    ["MuteBadSounds"] = true,
    ["MuteFlightSound"] = true,
    ["BulkSend"] = true,
    ["FastLoot"] = true,
    ["BigName"] = 16,
}


-- **HELPER FUNCTIONS**
-- TODO(GBA): This function is not safe. Please update.
function GBA:GetColorText(color, text)
    if color == "red" then
        color = "ffff0000"
    end
    if color == "green" then
        color = "ff00ff00"
    end
    return "|c"..color..text.."|r"
end

function GBA:ToggleChangeMsg(func)
    if E.db.GBA[func] then
        GBA:EnableMsg(func)
    else
        GBA:DisableMsg(func)
    end
end

function GBA:EnableMsg(func) print(GBA:GetColorText("green", "ElvUI_GBA: Enabled "..func)); end
function GBA:DisableMsg(func) print(GBA:GetColorText("red", "ElvUI_GBA: Disabled "..func)); end

local function CreateAndUpdateOG(...)
    local pred, unit, myIncomingHeal, otherIncomingHeal, absorb, healAbsorb, hasOverAbsorb, hasOverHealAbsorb, health, maxHealth = ...

    if pred.gbaOverlay == nil then
        pred.gbaOverlay = pred.health:CreateTexture(nil, "OVERLAY")
        pred.gbaOverlay:SetTexture("Interface\\RaidFrame\\Shield-Overlay", true, true)
        pred.gbaOverlay.tileSize = 32;
        pred.gbaOverlay:SetBlendMode("BLEND")
    end

    if pred.gbaGlow == nil then
        pred.gbaGlow = pred.health:CreateTexture(nil, "OVERLAY")
        pred.gbaGlow:SetAlpha(1)
        pred.gbaGlow:SetTexture("Interface\\RaidFrame\\Shield-Overshield")
        pred.gbaGlow:SetBlendMode("ADD")
        pred.gbaGlow:SetWidth(16)
        -- pred.gbaGlow:SetPoint('TOP')
        -- pred.gbaGlow:SetPoint('BOTTOM')
        -- pred.gbaGlow:SetPoint('LEFT', pred.health, 'RIGHT')
    end

    local overlay = pred.gbaOverlay
    local glow = pred.gbaGlow
    local absorbBar = pred.absorbBar

    overlay:ClearAllPoints()
    glow:ClearAllPoints()
    overlay:SetParent(pred.absorbBar)
    glow:SetParent(pred.absorbBar)

    if (absorb > maxHealth) then
        absorbBar = maxHealth
    end
    
    local totalWidth, totalHeight = pred.health:GetSize();
    local barSize = absorb / maxHealth * totalWidth;

    -- TODO(rxt): Check db.absorbStyle == WRAPPED
    if "WRAPPED" ~= "WRAPPED" then return end
    -- overlay:Hide()
    -- glow:Hide()
    if 0 == absorb then
        overlay:Hide()
        glow:Hide()
        return
    end
    if hasOverAbsorb then
        if health == maxHealth then
            overlay:SetPoint("TOPRIGHT", absorbBar, "TOPRIGHT", 0, 0);
            overlay:SetPoint("BOTTOMRIGHT", absorbBar, "BOTTOMRIGHT", 0, 0);
            overlay:SetWidth(barSize)
            overlay:SetTexCoord(0, barSize / overlay.tileSize, 0, totalHeight / overlay.tileSize);
            overlay:Show()
        else
            overlay:SetPoint("TOPRIGHT", absorbBar, "TOPRIGHT", 0, 0);
            overlay:SetPoint("BOTTOMRIGHT", absorbBar, "BOTTOMRIGHT", 0, 0);
            overlay:SetWidth(barSize)
            overlay:SetTexCoord(0, barSize / overlay.tileSize, 0, totalHeight / overlay.tileSize);
            overlay:Show()
        end
        glow:SetPoint("TOPLEFT", overlay, "TOPLEFT", -5, 1);
        glow:SetPoint("BOTTOMLEFT", overlay, "BOTTOMLEFT", -5, -1);
        glow:Show()
    else
        overlay:SetPoint("TOPLEFT", absorbBar, "TOPLEFT", 0, 0);
        overlay:SetPoint("BOTTOMLEFT", absorbBar, "BOTTOMLEFT", 0, 0);
        overlay:SetWidth(barSize)
        overlay:SetTexCoord(0, barSize / overlay.tileSize, 0, totalHeight / overlay.tileSize);
        overlay:Show()
        glow:Hide()
    end
end

-- **FEATURE FUNCTIONS**
function GBA:UpdateCombatDamageText() SetCVar("floatingCombatTextCombatDamage", E.db.GBA.CombatDamageText); end
function GBA:UpdateCombatHealingText() SetCVar("floatingCombatTextCombatHealing", E.db.GBA.CombatHealingText); end
function GBA:UpdateWorldTextScale() SetCVar("WorldTextScale", E.db.GBA.WorldTextScale); end

function GBA:UpdateSpellQueueWindow() SetCVar("SpellQueueWindow", E.db.GBA.SpellQueueWindow); end
function GBA:UpdateWeatherDensity() SetCVar("weatherDensity", E.db.GBA.WeatherDensity); end
function GBA:UpdateFfxGlow() SetCVar("ffxGlow", E.db.GBA.FfxGlow); end

function GBA:UpdateAutoFillDelete()
    if not E.db.GBA.AutoFillDelete then return end
    local f = function(box)
        if not E.db.GBA.AutoFillDelete then return end
        box.editBox:SetText(DELETE_ITEM_CONFIRM_STRING)
    end
    hooksecurefunc(StaticPopupDialogs.DELETE_GOOD_ITEM, 'OnShow', f)
end
function GBA:UpdateAutoGearCompare() SetCVar("alwaysCompareItems", E.db.GBA.AutoGearCompare); end
function GBA:UpdateFanhexie() SetCVar("overrideArchive", not E.db.GBA.Fanhexie); end
function GBA:UpdateProfanityFilter() SetCVar("profanityFilter", E.db.GBA.ProfanityFilter); end
function GBA:UpdateEnhancedReputation()
    if not E.db.GBA.EnhancedReputation then return end
    local NUM_FACTIONS_DISPLAYED = NUM_FACTIONS_DISPLAYED
    local REPUTATION_PROGRESS_FORMAT = REPUTATION_PROGRESS_FORMAT
    local function f(self, elapsed)
        ReputationFrame.paragonFramesPool:ReleaseAll()
        local factionOffset = FauxScrollFrame_GetOffset(ReputationListScrollFrame)
        for n=1, NUM_FACTIONS_DISPLAYED, 1 do
            local factionIndex = factionOffset + n
            local factionRow = _G["ReputationBar"..n]
            local factionBar = _G["ReputationBar"..n.."ReputationBar"]
            local factionStanding = _G["ReputationBar"..n.."ReputationBarFactionStanding"]
            if factionIndex <= GetNumFactions() then
                local name, _, _, _, _, _, _, _, _, _, _, _, _, factionID = GetFactionInfo(factionIndex)
                if factionID and C_Reputation.IsFactionParagon(factionID) then
                    local currentValue,threshold,rewardQuestID,hasRewardPending = C_Reputation.GetFactionParagonInfo(factionID)
                    local count = floor(currentValue / threshold)
                    if hasRewardPending then count = count - 1 end
                    factionRow.questID = rewardQuestID
                    if currentValue then
                        local value = mod(currentValue,threshold)
                        if hasRewardPending then
                            local paragonFrame = ReputationFrame.paragonFramesPool:Acquire()
                            paragonFrame.factionID = factionID
                            paragonFrame:SetPoint("RIGHT", factionRow, 11, 0)
                            paragonFrame.Glow:SetShown(true)
                            paragonFrame.Check:SetShown(true)
                            paragonFrame:Show()
                            value = value + threshold
                        end
                        factionBar:SetMinMaxValues(0, threshold)
                        factionBar:SetValue(value)
                        factionRow.rolloverText = format(REPUTATION_PROGRESS_FORMAT, BreakUpLargeNumbers(value), BreakUpLargeNumbers(threshold))
                        factionStanding:SetText(" "..BreakUpLargeNumbers(value).."/".."Paragon("..count..")")
                        factionRow.standingText = (" "..BreakUpLargeNumbers(value).."/".."Paragon("..count..")")
                        factionRow.rolloverText = nil                   
                        if factionIndex == GetSelectedFaction() and ReputationDetailFrame:IsShown() then
                            if count > 0 then
                                ReputationDetailFactionName:SetText(name.."|cffee80ee(Opened "..count.." times)|r")
                            end
                        end
                    end
                else
                    factionRow.questID = nil
                end
            else
                factionRow:Hide()
            end
        end
    end
    hooksecurefunc("ReputationFrame_Update", f)
end
function GBA:UpdateMuteBadSounds()
    if not E.db.GBA.MuteBadSounds then return end
    local stupidSoundIDs = {
        568250, -- sound/spells/goblin_weather_machine_rain.ogg
        568631, -- sound/spells/goblin_weather_machine_lightning.ogg
        568803, -- sound/spells/goblin_weather_machine_snow.ogg
        568904, -- sound/spells/goblin_weather_machine_cloudy.ogg
        -- All Toy Train sound/character/*/*train*.ogg
        539203,539219,539516,539730,541463,541601,541769,542017,542035,542206,630296,630298,636621,1306531,1313588
    }
    for _, soundID in pairs (stupidSoundIDs) do
       MuteSoundFile(soundID)
    end
end
function GBA:UpdateMuteFlightSound()
    if not E.db.GBA.MuteFlightSound then return end
    local flightStarted = 0
    local muting = false
    -- https://www.curseforge.com/wow/addons/silentflights
    local f = function(event, ...)
        if event == "TAXIMAP_CLOSED" then
            flightStarted = GetTimePreciseSec()
        end
        -- Note(GBA): If the toggle is flipped during flight, we make sure we don't mute sound.
        -- It's okay to turn on cuz people will notice it and know where to turn off if they don't want.
        -- It's kind of a tradeoff to avoid reload issue.
        if event == "PLAYER_CONTROL_LOST" and E.db.GBA.MuteFlightSound then
            if not C_Map.GetPlayerMapPosition(1550, "player") then
                return
            end
            if GetTimePreciseSec() - flightStarted < 1 then
                muting = true
                C_Timer.After(1, function()
                    SetCVar("Sound_EnableSFX", false)
                end)
            end
        end
        if event == "PLAYER_CONTROL_GAINED" then
            if muting then
                SetCVar("Sound_EnableSFX", true)
                muting = false
            end
        end
    end

    -- local eventFrame = CreateFrame("FRAME")
    self:RegisterEvent("TAXIMAP_CLOSED", f)
    self:RegisterEvent("PLAYER_CONTROL_LOST", f)
    self:RegisterEvent("PLAYER_CONTROL_GAINED", f)
    -- self:SetScript("OnEvent", f)
end
function GBA:UpdateBulkSend()
    if not E.db.GBA.BulkSend then return end
    local f = function(self, button)
        if not MailFrame:IsVisible()  or not E.db.GBA.BulkSend then return end
        if IsAltKeyDown() and button == "LeftButton" and self:GetParent():GetID() <= 4 then 
            local itemName = GetItemInfo(GetContainerItemID(self:GetParent():GetID(), self:GetID()))
            for b = 0, 4 do
                for s = 0, 32 do
                    l = GetContainerItemLink(b, s)
                    if l and l:find(itemName) then
                        UseContainerItem(b, s)
                    end
                end
            end
        end 
    end
    hooksecurefunc("ContainerFrameItemButton_OnModifiedClick", f)
end
function GBA:UpdateFastLoot()
    -- Note(GBA): It seems dumb to check twice but this can avoid reload.
    if not E.db.GBA.FastLoot then return end
    local lastLoot = 0
    local f = function(...)
        if not E.db.GBA.FastLoot or (GetTime() - lastLoot) <= 0 then return end
        if GetCVarBool("autoLootDefault") ~= IsModifiedClick("AUTOLOOTTOGGLE") then
            for i = GetNumLootItems(), 1, -1 do
                LootSlot(i)
            end
            lastLoot = GetTime()
        end
    end
    self:RegisterEvent("LOOT_READY", f)
end
function GBA:UpdateAdvnacedAbsorb()
    hooksecurefunc(ElvUF_Player_HealthBar.__owner.HealthPrediction, "PostUpdate", CreateAndUpdateOG)
    for i = 1, 5 do
        local t = _G["ElvUF_PartyGroup1UnitButton"..i.."_HealthBar"].__owner.HealthPrediction
        hooksecurefunc(t, "PostUpdate", CreateAndUpdateOG)
    end
end

function GBA:UpdateBigName()
    -- SystemFont_LargeNamePlate
    -- SystemFont_NamePlate
    -- SystemFont_LargeNamePlateFixed
    -- SystemFont_NamePlateFixed
    -- SystemFont_NamePlateCastBar
    if not E.db.GBA.BigName then return end
    local obj = SystemFont_NamePlate
    local fontName, _, fontFlags  = obj:GetFont()
    obj:SetFont(fontName, E.db.GBA.BigName, "OUTLINE")
    obj:SetShadowOffset(0, 0)
end

-- **Core Functions**
function GBA:UpdateAll()
    GBA:UpdateCombatDamageText()
    GBA:UpdateCombatHealingText()
    GBA:UpdateWorldTextScale()

    GBA:UpdateSpellQueueWindow()
    GBA:UpdateWeatherDensity()
    GBA:UpdateFfxGlow()

    GBA:UpdateAutoFillDelete()
    GBA:UpdateAutoGearCompare()
    GBA:UpdateFanhexie()
    GBA:UpdateProfanityFilter()
    GBA:UpdateEnhancedReputation()
    GBA:UpdateMuteBadSounds()
    GBA:UpdateMuteFlightSound()
    GBA:UpdateBulkSend()
    GBA:UpdateFastLoot()

    GBA:UpdateBigName()

    local f = CreateFrame("Frame")
    f:RegisterEvent("PLAYER_ENTERING_WORLD")
    f:SetScript("OnEvent", function(self, event) GBA:UpdateAdvnacedAbsorb() end)
end

-- This function inserts our GUI table into the ElvUI Config. You can read about AceConfig here: http://www.wowace.com/addons/ace3/pages/ace-config-3-0-options-tables/
function GBA:InsertOptions()
    E.Options.args.GBA = {
        type = "group",
        name = "GBA",
        get = function(info) return E.db.GBA[info[#info]] end,
        set = function(info, value) E.db.GBA[info[#info]] = value end,
        args = {
            ReloadIndicator = {
                order = 0,
                type = "header",
                name = "Options with * require reload to take effect.",
            },
            -- Combat Text
            CombatTabIndicator = {
                order = 10,
                type = "header",
                name = "Combat Text",
            },
            CombatDamageText = {
                order = 11,
                type = "toggle",
                name = "CombatDamageText",
                desc = "Show or hide combat damage text on targets.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateCombatDamageText(value); GBA:ToggleChangeMsg(info[#info]); end,
            },
            CombatHealingText = {
                order = 12,
                type = "toggle",
                name = "CombatHealingText",
                desc = "Show or hide combat healing text on targets.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateCombatHealingText(value); GBA:ToggleChangeMsg(info[#info]); end,
            },
            WorldTextScale = {
                order = 13,
                type = "range",
                name = "CombatTextScale",
                desc = "Increase the size of combat text and some related world text.",
                min = 1,
                max = 2.5,
                step = 0.1,
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateWorldTextScale(value); end,
            },
            -- CVars
            CVarSettingIndicator = {
                order = 20,
                type = "header",
                name = "Common CVars",
            },
            SpellQueueWindow = {
                order = 21,
                type = "range",
                name = "SpellQueueWindow",
                min = 0,
                max = 400,
                step = 1,
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateSpellQueueWindow(value); end,
            },
            WeatherDensity = {
                order = 22,
                type = "range",
                name = "WeatherDensity",
                desc = "0: disable, 1~3 for different weather density.",
                min = 0,
                max = 3,
                step = 1,
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateWeatherDensity(value); end,
            },
            FfxGlow = {
                order = 23,
                type = "toggle",
                name = "ffxGlow",
                desc = "Enable/disable full screen glow. e.g. low hp.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateFfxGlow(value); end,
            },
            -- Toolkit
            ToolkitIndicator = {
                order = 30,
                type = "header",
                name = "Game Enhancement",
            },
            AutoFillDelete = {
                order = 31,
                type = "toggle",
                name = "AutoFillDelete",
                desc = "Auto Fill DELETE when destorying items.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateAutoFillDelete(); GBA:ToggleChangeMsg(info[#info]); end,
            },
            AutoGearCompare = {
                order = 32,
                type = "toggle",
                name = "AutoGearCompare",
                desc = "Disabling Auto Gear Compare may improve ElvUI performance.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateAutoGearCompare(); GBA:ToggleChangeMsg(info[#info]); end,
            },
            Fanhexie = {
                order = 33,
                type = "toggle",
                name = "反和谐*",
                desc = "An option for zhCN client only.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateFanhexie(); GBA:ToggleChangeMsg(info[#info]); end,
            },
            ProfanityFilter = {
                order = 34,
                type = "toggle",
                name = "Language Filter",
                desc = "Enable or disable profanity filter.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateProfanityFilter(); GBA:ToggleChangeMsg(info[#info]); end,
            },
            EnhancedReputation = {
                order = 35,
                type = "toggle",
                name = "AdvancedParagon*",
                desc = "Show advanced paragon information on the reputation page.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:ToggleChangeMsg(info[#info]); end,
            },
            MuteBadSounds = {
                order = 36,
                type = "toggle",
                name = "MuteBadSounds*",
                desc = "Mute annoying sounds like toy train, etc.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:ToggleChangeMsg(info[#info]); end,
            },
            MuteFlightSound = {
                order = 37,
                type = "toggle",
                name = "MuteFlightSound*",
                desc = "A workaround to mute annoying flight sound between oribos and other lands.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateMuteFlightSound(); GBA:ToggleChangeMsg(info[#info]); end,
            },
            BulkSend = {
                order = 38,
                type = "toggle",
                name = "BulkSend",
                desc = "Enabling this will allow you add all other same items to mail attachment by ALT+LeftClick on an item.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateBulkSend(); GBA:ToggleChangeMsg(info[#info]); end,
            },
            FastLoot = {
                order = 39,
                type = "toggle",
                name = "FastLoot",
                desc = "Make auto loot faster.",
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateFastLoot(); GBA:ToggleChangeMsg(info[#info]); end,
            },
            BigName = {
                order = 40,
                type = "range",
                name = "BigName",
                desc = "Increase the friendly name size in instance.",
                min = 10,
                max = 30,
                step = 1,
                set = function(info, value) E.db.GBA[info[#info]] = value; GBA:UpdateBigName(); end,
            },
            -- End
            EndIndicator = {
                order = 200,
                type = "header",
                name = string.format("ElvUI_GBA %s by GBA", addonVersion),
            }, 
        },
    }
end

function GBA:Initialize()
    -- Register plugin so options are properly inserted when config is loaded
    EP:RegisterPlugin(addonName, GBA.InsertOptions)
    -- Make sure everything remains good on each reload.
    GBA:UpdateAll()
end

E:RegisterModule(GBA:GetName()) --Register the module with ElvUI. ElvUI will now call GBA:Initialize() when ElvUI is ready to load our plugin.


-- **SLASH COMMAND**
local function GBACommands(msg, editbox)
    if msg == "reset" then
        for k in pairs (E.db.GBA) do
            E.db.GBA[k] = nil
        end
        E:StaticPopup_Show('PRIVATE_RL')
    elseif msg == "config" then
        GBA:DumpTable(E.db.GBA)
    elseif msg == "debug" then
        GBA:DumpTable(addonTable)
    else
        print("GBA: hi~")
    end
end

SLASH_GBA1, SLASH_GBA2 = "/gba", "/elvui_gba"
SlashCmdList["GBA"] = GBACommands